# :coding: utf-8

import argparse
import json
import os
import sys


def display_python_mapping():
    """Display Python version information mapping.

    The information is displayed as a :term:`JSON` encoded mapping so that it
    can easily be retrieved via a subprocess.

    :return: None

    Example::

        >>> display_python_mapping()

        {
            "identifier": "2.7",
            "request": "python >= 2.7, < 2.8",
            "library-path": "lib/python2.7/site-packages"
        }


    """
    python_version = sys.version_info

    name = "python{major}.{minor}".format(
        major=sys.version_info.major,
        minor=sys.version_info.minor,
    )

    mapping = {
        "identifier": "{major}.{minor}".format(
            major=python_version.major,
            minor=python_version.minor,
        ),
        "request": "python >= {major}.{minor}, < {major}.{next_minor}".format(
            major=python_version.major,
            minor=python_version.minor,
            next_minor=python_version.minor + 1
        ),
        "library-path": os.path.join("lib", name, "site-packages")
    }

    print(json.dumps(mapping, sort_keys=True, indent=4))


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        prog="python-info",
        description="Query information about a Python environment",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    namespace = parser.parse_args()

    sys.exit(display_python_mapping())



# :coding: utf-8

import json
import re
import os
import textwrap

import pytest
import wiz.utility

import qip.command
import qip.system
from qip.environ import PYTHON_INFO_SCRIPT
from qip.package import PACKAGE_INFO_SCRIPT

#: Context mapping mocked for Python 2.7
PYTHON_MAPPING_27 = {
    "identifier": "2.7",
    "library-path": "lib/python2.7/site-packages",
    "request": "python >= 2.7, < 2.8"
}

#: Collection of all output folder .
PACKAGE_OUTPUT_PATH = {
    "foo": os.path.join("Foo", "Foo-1.2.0-py27"),
    "bar": os.path.join("Bar", "Bar-0.1.0-py27-centos7"),
    "bim": os.path.join("BIM", "BIM-3.6.2-py27"),
}

#: Collection of all packages dependency mapping.
PACKAGE_DEPENDENCY_MAPPING = {
    "foo": {
        "package": {
            "installed_version": "1.2.0",
            "key": "foo",
            "module_name": "foo",
            "package_name": "Foo"
        },
        "requirements": [
            "bim >= 3.4, < 5"
        ]
    },
    "bar": {
        "package": {
            "installed_version": "0.1.0",
            "key": "bar",
            "module_name": "bar",
            "package_name": "Bar"
        },
        "requirements": [
            "foo"
        ]
    },
    "bim": {
        "package": {
            "installed_version": "3.6.2",
            "key": "bim",
            "module_name": "bim",
            "package_name": "BIM"
        }
    },
}

#: Collection of metadata per package name.
METADATA_MAPPING = {
    "foo": textwrap.dedent(
        """
        Summary: Foo Python Package.
        Location: /path/to/foo
        Entry-points:
          [console_scripts]
          foo = foo.__main__:main
        """
    ),
    "bar": textwrap.dedent(
        """
        Summary: Bar Python Package.
        Location: /path/to/bar
        Classifiers:
          Operating System :: Unix
        """
    ),
    "bim": textwrap.dedent(
        """
        Summary: Bim Python Package.
        Location: /path/to/bim
        """
    ),
}


@pytest.fixture(autouse=True)
def mock_commands(mocker):
    """Mock all subprocesses used for installing a package with Python 2.7."""

    def _command_execute(command, *args, **kwargs):
        """Mock :func:`qip.command.execute` depending on context."""
        if command.startswith("python {}".format(PYTHON_INFO_SCRIPT)):
            return json.dumps(PYTHON_MAPPING_27)

        elif command.startswith("python -m pip install"):
            request = re.search(r"'(.+)'", command).group(1)
            output = re.search(r"--prefix ([^ ]+)", command).group(1)
            requirement = wiz.utility.get_requirement(request)
            os.makedirs(
                os.path.join(output, PACKAGE_OUTPUT_PATH[requirement.name])
            )
            return "Installing collected packages: {}".format(requirement.name)

        elif command.startswith("python {}".format(PACKAGE_INFO_SCRIPT)):
            identifier = command.rsplit(" ", 1)[-1]
            return json.dumps(PACKAGE_DEPENDENCY_MAPPING[identifier])

        elif command.startswith("python -m pip show"):
            name = re.search(r"'(.+)'", command).group(1)
            return METADATA_MAPPING[name]

    mocker.patch.object(qip.command, "execute", _command_execute)


@pytest.fixture(autouse=True)
def mock_system(mocker):
    """Mock system mapping"""
    mocker.patch.object(
        qip.system, "query",
        return_value={
            "platform": "linux",
            "arch": "x86_64",
            "os": {
                "name": "centos",
                "major_version": 7
            }
        }
    )



# :coding: utf-8

import subprocess

import pytest

import qip.command


@pytest.fixture()
def mocked_subprocess(mocker):
    """Return mocked subprocess."""
    _mocked_command = mocker.patch.object(subprocess, "Popen")
    return _mocked_command


@pytest.fixture()
def mocked_process(mocker):
    """Return mocked process."""
    return mocker.Mock()


def test_execute_verbose(mocked_subprocess, mocked_process, logger):
    """Execute a command verbose."""
    mocked_subprocess.return_value = mocked_process
    mocked_process.poll.side_effect = [None, None, None, False]
    mocked_process.stdout.readline.side_effect = [
        b"line one\n", b"line two\n", b"line three\n"
    ]
    mocked_process.stderr.readlines.return_value = []

    command = "pip install foo"
    output = qip.command.execute(command, {})

    assert logger.debug.call_count == 4
    logger.debug.assert_any_call("pip install foo")
    logger.debug.assert_any_call("line one")
    logger.debug.assert_any_call("line two")
    logger.debug.assert_any_call("line three")

    assert output == "line one\nline two\nline three\n"


def test_execute_quiet(mocked_subprocess, mocked_process, capsys):
    """Execute a command quiet."""
    mocked_subprocess.return_value = mocked_process
    mocked_process.communicate.return_value = (
        b"line one\nline two\nline three\n", b""
    )

    command = "pip install foo"
    output = qip.command.execute(command, {}, quiet=True)

    stdout_message, _ = capsys.readouterr()
    assert stdout_message == ""
    assert output == "line one\nline two\nline three\n"


def test_execute_stderr(mocked_subprocess, mocked_process):
    """Fail to execute a command."""
    mocked_subprocess.return_value = mocked_process
    mocked_process.communicate.return_value = (b"", b"stderr")
    command = "pip install foo"

    with pytest.raises(RuntimeError) as error:
        qip.command.execute(command, {}, quiet=True)
    assert str(error.value) == "stderr"



# :coding: utf-8

import sys
import os
import re
import importlib

import pytest
from six.moves import reload_module
from click.testing import CliRunner
import wiz
import wiz.config

import qip.command_line


#: Record name of python library path for convenience.
PY_LIB = "python{}.{}".format(sys.version_info.major, sys.version_info.minor)


@pytest.fixture(autouse=True)
def reset_configuration(mocker):
    """Ensure that no personal configuration is fetched during tests."""
    function = os.path.expanduser
    mocker.patch.object(os.path, "expanduser", return_value="__HOME__")

    # Reset configuration.
    qip.command_line._CONFIG = wiz.config.fetch(refresh=True)
    reload_module(qip.command_line)

    # Reset mock for 'os.path.expanduser' to prevent messing with the tests.
    mocker.patch.object(os.path, "expanduser", function)


def test_install_numpy(temporary_directory, logger):
    """Install numpy.

    Install numpy version within 1.16.6 and 1.19.2 which covers all Python
    version supported by Qip. numpy doesn't have any dependencies.

    """
    packages_path = os.path.join(temporary_directory, "packages")
    definitions_path = os.path.join(temporary_directory, "definitions")

    runner = CliRunner()
    result = runner.invoke(
        qip.command_line.main, [
            "install",
            "numpy >= 1.16.6, <= 1.19.2",
            "-o", packages_path,
            "-d", definitions_path,
        ]
    )
    assert not result.exception
    assert result.exit_code == 0

    # Check log.
    logger.info.assert_called()
    logger.warning.assert_not_called()
    logger.error.assert_not_called()

    # Check definitions installed.
    definitions = os.listdir(definitions_path)
    assert len(definitions) == 1
    assert re.match(r"^library-numpy-1..*.json$", definitions[0])
    path = os.path.join(definitions_path, definitions[0])
    assert os.path.isfile(path)

    # Ensure that definition is correct.
    definition = wiz.load_definition(path)
    assert definition.qualified_identifier == "library::numpy"
    assert definition.install_root == packages_path
    assert definition.system is not None
    assert definition.requirements == []
    assert definition.environ == {
        "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
    }
    assert len(definition.variants) == 1
    assert len(definition.variants[0].requirements) == 1
    assert definition.variants[0].requirements[0].name == "python"

    # Check package installed.
    assert len(os.listdir(packages_path)) == 1
    path = os.path.join(packages_path, "numpy")
    assert os.path.isdir(path)

    versions = os.listdir(path)
    assert len(versions) == 1
    assert re.match(r"^numpy-1..*", versions[0])
    lib_path = os.path.join(path, versions[0], "lib", PY_LIB, "site-packages")
    module_path = os.path.join(lib_path, "numpy")
    assert os.path.isdir(lib_path)
    assert os.path.isdir(module_path)

    # Check that module can be imported.
    sys.path.insert(0, lib_path)
    importlib.import_module("numpy")
    del sys.path[0]


def test_install_numpy_several_versions(temporary_directory, logger):
    """Install numpy.

    Like the previous test, install numpy version within 1.16.6 and 1.19.2 which
    covers all Python version supported by Qip. Then install numpy while
    excluding the latest versions supported by 2.7 and 3+ to force another
    version to get installed.

    """
    packages_path = os.path.join(temporary_directory, "packages")
    definitions_path = os.path.join(temporary_directory, "definitions")

    runner = CliRunner()
    result = runner.invoke(
        qip.command_line.main, [
            "install",
            "numpy >= 1.16.6, <= 1.19.2",
            "numpy != 1.16.6, < 1.19.2.",
            "-o", packages_path,
            "-d", definitions_path,
        ]
    )
    assert not result.exception
    assert result.exit_code == 0

    # Check log.
    logger.info.assert_called()
    logger.warning.assert_not_called()
    logger.error.assert_not_called()

    # Check definitions installed.
    definitions = os.listdir(definitions_path)
    assert len(definitions) == 2
    assert definitions[0] != definitions[1]

    assert re.match(r"^library-numpy-1..*.json$", definitions[0])
    path = os.path.join(definitions_path, definitions[0])
    assert os.path.isfile(path)
    definition1 = wiz.load_definition(path)

    assert re.match(r"^library-numpy-1..*.json$", definitions[1])
    path = os.path.join(definitions_path, definitions[1])
    assert os.path.isfile(path)
    definition2 = wiz.load_definition(path)

    # Ensure that definitions are correct.
    for definition in [definition1, definition2]:
        assert definition.qualified_identifier == "library::numpy"
        assert definition.install_root == packages_path
        assert definition.system is not None
        assert definition.requirements == []
        assert definition.environ == {
            "PYTHONPATH": "${INSTALL_LOCATION}:${PYTHONPATH}"
        }
        assert len(definition.variants) == 1
        assert len(definition.variants[0].requirements) == 1
        assert definition.variants[0].requirements[0].name == "python"

    assert definition1.version != definition2.version

    # Check packages installed.
    assert len(os.listdir(packages_path)) == 1
    path = os.path.join(packages_path, "numpy")
    assert os.path.isdir(path)

    versions = os.listdir(path)
    assert len(versions) == 2

    for version in versions:
        assert re.match(r"^numpy-1..*", version)
        lib_path = os.path.join(path, version, "lib", PY_LIB, "site-packages")
        module_path = os.path.join(lib_path, "numpy")
        assert os.path.isdir(lib_path)
        assert os.path.isdir(module_path)

        # Check that module can be imported.
        sys.path.insert(0, lib_path)
        importlib.import_module("numpy")
        del sys.path[0]



# :coding: utf-8

import os
import sys
import tempfile
import shutil

import six.moves
import click
import wiz
import wiz.filesystem

import qip.definition
import qip.package
import qip.environ
import qip.logging

from qip._version import __version__


def install(
    requests, output_path, definition_path=None, overwrite=False,
    no_dependencies=False, editable_mode=False, python_target=sys.executable,
    registry_paths=None, update_existing_definitions=False,
    continue_on_error=False
):
    """Install packages to *output_path* from *requests*.

    :param requests: List of package requests to be installed.

        A request can be one of::

            ["."]
            ["/path/to/foo/"]
            ["foo", "bar"]
            ["foo==0.1.0"]
            ["foo >= 7, < 8"]
            ["git@gitlab:rnd/foo.git"]
            ["git@gitlab:rnd/foo.git@0.1.0"]
            ["git@gitlab:rnd/foo.git@dev"]

    :param output_path: Destination path for Python packages installation.

    :param definition_path: Destination path for :term:`Wiz` definitions
        automatically created for Python packages installed. Default is None,
        which means that no :term:`Wiz` definitions are created.

    :param overwrite: Indicate whether packages already installed in destination
        path and corresponding :term:`Wiz` definitions should be overwritten. If
        None, a user confirmation will be prompted. Default is False.

    :param no_dependencies: Indicate whether package dependencies should be
        skipped. Default is False.

    :param editable_mode: Indicate whether the Python package location should
        target the source installation package. Default is False.

    :param python_target: Target a specific Python version via a Wiz request or
        a path to a Python executable (e.g. "python==2.7.*" or
        "/path/to/bin/python"). Default is the path to the current Python
        executable.

    :param registry_paths: List of :term:`Wiz` registry paths to consider when
        fetching existing definitions to update or skip.

    :param update_existing_definitions: Indicate whether variants from existing
        definitions should be used when exporting new definitions.

    :param continue_on_error: Indicate whether installation process should
        continue if a package cannot be installed. Default is False.

    :raises: :exc:`RuntimeError` if a package cannot be installed and
        *continue_on_error* is set to False.

    :return: Boolean value indicating whether packages were installed.

    """
    logger = qip.logging.Logger(__name__ + ".install")

    wiz.filesystem.ensure_directory(output_path)
    if definition_path is not None:
        wiz.filesystem.ensure_directory(definition_path)

    # Fetch definition mapping to determining whether a package should be
    # skipped or updated.
    definition_mapping = wiz.fetch_definition_mapping(registry_paths or [])

    # Setup temporary folder for package installation.
    cache_path = tempfile.mkdtemp()
    package_path = tempfile.mkdtemp()

    # Record requests and package installed to prevent duplications.
    installed_packages = set()
    installed_requests = set()

    # Record packages skipped.
    skipped_packages = set()

    # Fill up queue with requirements extracted from requests.
    queue = six.moves.queue.Queue()

    try:
        # Fetch environment mapping and installation path.
        context_mapping = fetch_context_mapping(package_path, python_target)
        library_path = context_mapping["environ"]["PYTHONPATH"]

        for request in requests:
            queue.put((request, None, editable_mode))

        while not queue.empty():
            request, parent_identifier, _editable_mode = queue.get()
            if request in installed_requests:
                continue

            # Clean up before installation.
            shutil.rmtree(package_path)
            wiz.filesystem.ensure_directory(package_path)

            # Needed for the editable mode.
            wiz.filesystem.ensure_directory(library_path)

            package_mapping, overwrite = _install(
                request, output_path, context_mapping, definition_mapping,
                package_path, cache_path, installed_packages,
                definition_path=definition_path,
                overwrite=overwrite,
                editable_mode=_editable_mode,
                update_existing_definitions=update_existing_definitions,
                parent_identifier=parent_identifier,
                continue_on_error=continue_on_error
            )
            if package_mapping is None:
                continue

            installed_packages.add(package_mapping["identifier"])
            installed_requests.add(request)

            # Indicate if package was skipped.
            if package_mapping.get("skipped", False):
                skipped_packages.add(package_mapping["identifier"])

            # Fill up queue with requirements extracted from package
            # dependencies.
            if not no_dependencies:
                for request in package_mapping.get("requirements", []):
                    queue.put((request, package_mapping["identifier"], False))

    finally:
        shutil.rmtree(package_path)
        shutil.rmtree(cache_path)

    # Sort and filter packages installed.
    installed = sorted(
        (p for p in installed_packages if p not in skipped_packages),
        key=lambda _id: _id.lower()
    )

    if len(installed):
        logger.info("Packages installed: {}".format(", ".join(installed)))

    return len(installed_packages) > 0


def _install(
    request, output_path, context_mapping, definition_mapping,
    package_path, cache_path, installed_packages, definition_path=None,
    overwrite=False, update_existing_definitions=False, editable_mode=False,
    continue_on_error=False, parent_identifier=None,
):
    """Install single package to *output_path* from *request*.

    :param request: Package requests to be installed.

    :param output_path: root destination path for Python packages installation.

    :param context_mapping: Mapping containing environment and python mapping
        as returned by :func:`fetch_context_mapping`

    :param definition_mapping: mapping regrouping all available definitions as
        returned by :func:`wiz.fetch_definition_mapping`.

    :param package_path: Temporary path to install package from :term:`Pip`.

    :param cache_path: Temporary directory for the :term:`Pip` cache.

    :param installed_packages: Set grouping all Python package identifiers
        already installed to skip current installation if necessary.

    :param definition_path: Destination path for :term:`Wiz` definition
        automatically created for Python package installed. Default is None,
        which means that no :term:`Wiz` definition is created.

    :param overwrite: Indicate whether package already installed in destination
        path and corresponding :term:`Wiz` definition should be overwritten. If
        None, a user confirmation will be prompted. Default is False.

    :param editable_mode: Indicate whether the Python package location should
        target the source installation package. Default is False.

    :param update_existing_definitions: Indicate whether variants from existing
        definitions should be used when exporting new definitions.

    :param continue_on_error: Indicate whether installation process should
        continue if a package cannot be installed. Default is False.

    :param parent_identifier: Indicate Python package which triggered the
        *request*. Default is None.

    :raises: :exc:`RuntimeError` if a package cannot be installed and
        *continue_on_error* is set to False.

    :return: tuple with package mapping installed (or None is the installation
        couldn't be processed), and one boolean value indicating a new value for
        the *overwrite* option.

    .. note::

        Package mapping returned will have an additional "skipped" keyword
        set as True if existing definition has been found in default registries.


    """
    logger = qip.logging.Logger(__name__ + "._install")

    try:
        package_mapping = qip.package.install(
            request, package_path, context_mapping, cache_path,
            editable_mode=editable_mode
        )

    except RuntimeError as error:
        if not continue_on_error:
            raise

        prompt = "Request '{}' has failed".format(request)
        if parent_identifier is not None:
            prompt += " [from '{}']".format(parent_identifier)

        logger.error("{}:\n{}".format(prompt, error))
        return None, overwrite

    if package_mapping["identifier"] in installed_packages:
        return None, overwrite

    # Attempt to fetch custom definition from package and existing definition
    # from definition mapping.
    custom_definition = None
    existing_definition = None

    if definition_path is not None:
        custom_definition = qip.definition.fetch_custom(package_mapping)
        existing_definition = qip.definition.fetch_existing(
            package_mapping, definition_mapping,
            namespace=getattr(custom_definition, "namespace", None),
        )

        if not editable_mode and _skip_install(
            existing_definition, package_mapping, definition_path
        ):
            logger.warning(
                "Skip '{0[key]}[{0[python][identifier]}]=={0[version]}' "
                "which already exists in Wiz registries."
                .format(package_mapping)
            )
            package_mapping["skipped"] = True
            return package_mapping, overwrite

    prompt = "Requested '{}'".format(request)
    if parent_identifier is not None:
        prompt += " [from '{}']".format(parent_identifier)
    logger.info(prompt)

    # Install package to destination.
    skipped, overwrite = copy_to_destination(
        package_mapping, package_path, output_path,
        overwrite=overwrite
    )

    # Extract a wiz definition is requested.
    if not skipped and definition_path is not None:
        qip.definition.export(
            definition_path, package_mapping, output_path,
            editable_mode=editable_mode,
            existing_definition=(
                existing_definition if update_existing_definitions else None
            ),
            custom_definition=custom_definition
        )

    package_mapping["skipped"] = skipped
    return package_mapping, overwrite


def _skip_install(existing_definition, package_mapping, definition_path):
    """Indicate whether existing definition mandates installation to be skipped.

    Skip package installation if existing definition:

    1. Is not None;
    2. Doesn't have a variant for the current Python version used;
    3. Is not stored in registry where installation process will export new
      definitions.

    Point 3 will be handled by :func:`copy_to_destination`.

    """
    python_version = package_mapping["python"]["identifier"]

    if existing_definition is None:
        return False

    if not any(
        variant.identifier == python_version for variant
        in existing_definition.variants
    ):
        return False

    return existing_definition.registry_path != definition_path


def copy_to_destination(
    mapping, source_path, destination_path, overwrite=False
):
    """Copy package from *source_path* to *destination_path*.

    :param mapping: mapping of the python package built as returned by
        :func:`qip.package.install`.

    :param source_path: path where the package was built.

    :param destination_path: path to install to.

    :param overwrite: indicate whether packages already installed should be
        overwritten. If None, a user confirmation will be prompted. Default is
        False.

    :return: tuple with one boolean value indicating whether the copy has been
        skipped and one indicating a new value for the *overwrite* option.

    """
    logger = qip.logging.Logger(__name__ + ".copy_to_destination")

    identifier = mapping["identifier"]
    target = os.path.join(destination_path, mapping["target"])

    # By default, future overwrite request will the same as the present one.
    overwrite_next = overwrite

    if os.path.isdir(target):
        if overwrite is None:
            overwrite, overwrite_next = _confirm_overwrite(identifier)

        if overwrite:
            logger.warning(
                "Overwrite '{}' which is already installed.".format(identifier)
            )
            shutil.rmtree(target)

        else:
            logger.warning(
                "Skip '{}' which is already installed.".format(identifier)
            )

            return True, overwrite_next

    wiz.filesystem.ensure_directory(os.path.dirname(target))
    shutil.copytree(source_path, target)
    logger.debug("Source copied to '{}'".format(target))

    logger.info("\tInstalled '{}'.".format(identifier))

    return False, overwrite_next


def _confirm_overwrite(identifier):
    """Package overwrite confirmation prompt for *identifier*

    Return a tuple with one boolean value indicating whether *identifier* should
    be overwritten and one boolean indicating whether future packages should
    also be overwritten.

    """
    message = (
        "Overwrite '{}'? ([y]es, [n]o, [ya] yes to all, [na] no to all)"
        .format(identifier)
    )

    answer = click.prompt(
        message,
        type=click.Choice(["y", "n", "ya", "na"]),
        default="n",
        show_choices=False,
        show_default=False
    )

    overwrite = answer[0] == "y"
    overwrite_next = overwrite if "a" in answer else None
    return overwrite, overwrite_next


def fetch_context_mapping(path, python_target):
    """Return context mapping containing environment and python mapping.

    :param path: path where python package has been installed.

    :param python_target: Target a specific Python version via a Wiz request or
        a path to a Python executable (e.g. "python==2.7.*" or
        "/path/to/bin/python").

    :return: Context mapping.

        It should be in the form of::

            {
                "environ": {
                    "PATH": "/path/to/bin",
                    "PYTHONPATH": "/path/to/lib/python2.7/site-packages",
                },
                "python": {
                    "identifier": "2.7",
                    "request": "python >= 2.7, < 2.8",
                    "installation-target": "lib/python2.7/site-packages"
                }
            }

    """
    environ_mapping = qip.environ.fetch(
        python_target, mapping={"PYTHONWARNINGS": "ignore"}
    )

    # Fetch Python version mapping from environment
    python_mapping = qip.environ.fetch_python_mapping(environ_mapping)

    # Compute the installation path and add it to PYTHONPATH
    install_path = os.path.join(path, python_mapping["library-path"])
    environ_mapping["PYTHONPATH"] = install_path

    return {
        "environ": environ_mapping,
        "python": python_mapping
    }



# :coding: utf-8

import tempfile

import pytest
import wiz

import qip.environ
import qip.command


@pytest.fixture()
def mocked_mkdtemp(mocker):
    """Return mocked 'tempfile.mkdtemp' function"""
    return mocker.patch.object(tempfile, "mkdtemp")


@pytest.fixture()
def mocked_wiz_resolve_context(mocker):
    """Return mocked 'wiz.resolve_context' function"""
    return mocker.patch.object(wiz, "resolve_context")


@pytest.fixture()
def mocked_command_execute(mocker):
    """Return mocked command execute."""
    return mocker.patch.object(qip.command, "execute")


def test_fetch_environ(mocked_mkdtemp, mocked_wiz_resolve_context):
    """Fetch and return environment mapping."""
    mocked_wiz_resolve_context.return_value = {"environ": "__ENVIRON__"}

    environ = qip.environ.fetch("python==2.7.*")
    assert environ == "__ENVIRON__"

    mocked_wiz_resolve_context.assert_called_once_with(
        ["python==2.7.*"], environ_mapping={}
    )
    mocked_mkdtemp.assert_not_called()


def test_fetch_environ_with_mapping(mocked_mkdtemp, mocked_wiz_resolve_context):
    """Fetch and return environment mapping with initial mapping."""
    mocked_wiz_resolve_context.return_value = {"environ": "__ENVIRON__"}

    environ = qip.environ.fetch("python==2.7.*", mapping="__INITIAL_MAPPING__")
    assert environ == "__ENVIRON__"

    mocked_wiz_resolve_context.assert_called_once_with(
        ["python==2.7.*"], environ_mapping="__INITIAL_MAPPING__"
    )
    mocked_mkdtemp.assert_not_called()


def test_fetch_environ_with_python_path(
    temporary_directory, mocked_mkdtemp, mocked_wiz_resolve_context
):
    """Fetch and return environment mapping with python path."""
    mocked_wiz_resolve_context.return_value = {"environ": "__ENVIRON__"}
    mocked_mkdtemp.return_value = temporary_directory

    environ = qip.environ.fetch("/bin/python")
    assert environ == {"PATH": "{}:${{PATH}}".format(temporary_directory)}

    mocked_wiz_resolve_context.assert_not_called()


def test_fetch_python_mapping(mocked_command_execute):
    """Fetch and return Python mapping."""
    mocked_command_execute.return_value = "{\"python\": \"__MAPPING__\"}"

    mapping = qip.environ.fetch_python_mapping("__ENVIRON__")
    assert mapping == {"python": "__MAPPING__"}


def test_fetch_python_mapping_error(mocked_command_execute):
    """Fail to fetch and return Python mapping."""
    mocked_command_execute.return_value = "{{{"

    with pytest.raises(RuntimeError):
        qip.environ.fetch_python_mapping("__ENVIRON__")



# :coding: utf-8

import shlex
import subprocess

import qip.logging


def execute(command, environ_mapping, quiet=False):
    """Execute *command* within *environ_mapping*.

    :param command: command to execute.

        It should be in the form of::

            "python -m pip install foo"

    :param environ_mapping: mapping with all environment variables that
        must be set for the *command* to run properly.
    :param quiet: indicate whether output lines should be hidden.
        Default is False.

    :raise RuntimeError: if command execution fails

    :return: Command output.

    """
    logger = qip.logging.Logger(__name__ + ".execute")
    logger.debug(command)

    process = subprocess.Popen(
        shlex.split(command),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        env=environ_mapping
    )

    if not quiet:
        output = ""

        lines_iterator = iter(process.stdout.readline, b"")
        while process.poll() is None:
            for line in lines_iterator:
                output += line.decode("utf-8")
                _line = line.rstrip().decode("utf-8")
                logger.debug(_line)

        stderr = b"\n".join(process.stderr.readlines()).decode("utf-8")
    else:
        output, stderr = process.communicate()

        # Decode process output for display
        output = output.decode("utf-8")
        stderr = stderr.decode("utf-8")

    if len(stderr):
        raise RuntimeError(stderr)

    return output
